#
#  HTMLTagDict.py
#  JunkMatcher
#
#  Created by Benjamin Han on 2/1/05.
#  Copyright (c) 2005 Benjamin Han. All rights reserved.
#

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#!/usr/bin/env python

from consts import *
from utilities import *


class HTMLTagDict (dict):
    """Reads in a text file (see htmlTags.txt for an example) and stores all
    valid tags: for each tag the attributes that can take an URL as a value are also
    stored as a set."""
    def __init__ (self, fn):
        self.fn = fn
        self.load()

    def load (self):
        self.clear()
        for line in filter(lambda l: len(l), map(lambda l: l.split('#')[0].strip(),
                                                 openFile(self.fn))):
            comps = line.split(' ')
            tag = comps[0]
            if len(comps) > 1:
                if comps[1] == '*':
                    self[tag] = '*'  # URLs can freely appear in this tag
                else:
                    self[tag] = sets.Set(comps[1:])
            else:
                self[tag] = None

    def setTag (self, tag, attrs):
        """Sets (add new or update existing) an HTML tag with attributes in 'attrs'
        (a space-separated and stripped string)."""
        if len(attrs) == 0:
            attrs = None
        elif attrs != '*':
            attrs = attrs.split(' ')
        
        self[tag] = attrs

    def removeTag (self, tag):
        try:
            del self[tag]
        except:
            pass

    def writeToFile (self):
        # don't reuse __str__ - otherwise unicode will raise exceptions!
        itemList = self.items()
        itemList.sort()
        sList = []
        for k, v in itemList:
            if v:
                sList.append('%s %s' % (k, ' '.join(v)))
            else:
                sList.append('%s' % k)
                
        openFile(self.fn, 'w').write('\n'.join(sList))

    def __str__ (self):
        itemList = self.items()
        itemList.sort()
        sList = []
        for k, v in itemList:
            if v:
                sList.append('%s %s' % (k, ' '.join(v)))
            else:
                sList.append('%s' % k)

        return '\n'.join(sList)


if __name__ == '__main__':
    htmlTagDict = HTMLTagDict('%shtmlTags' % CONF_PATH)
    print htmlTagDict
