#
#  backupOldSettings.py
#  JunkMatcher
#
#  Created by Benjamin Han on 3/11/05.
#  Copyright (c) 2005 Benjamin Han. All rights reserved.
#

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import sys, shutil

sys.path.insert(0, '../Engine')

from consts import *
from utilities import *

patternFiles = [('junkContent.txt', 'Patterns for message body', 'body'),
                ('junkFilename.txt', 'Patterns for attachment filenames', 'filenames'),
                ('junkHTML.txt', 'Patterns for message rendering', 'rendering'),
                ('junkHeaders.txt', 'Patterns for message headers', 'headers'),
                ('junkSender.txt', 'Patterns for senders', 'sender'),
                ('junkSubj.txt', 'Patterns for message subjects', 'subject'),
                ('metaPatterns.txt', 'Meta patterns', None),
                ('safeIP.txt', 'Patterns for Safe IPs', None),
                ('safeRecipients.txt', 'Patterns of your email addresses', None)]

htmlTemplate = '''<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
<html>
<head>
<meta http-equiv="Content-Type" content="text/html; charset=utf-8">
<link rel="stylesheet" href=".base.css" type="text/css">
<title>%s</title>
</head>
<body>
<h2>%s</h2>
<p>%s</p>
%s
</body>
</html>
'''


def readPatternsFromFN (fn):
    try:
        f = openFile(fn, 'r', 'macroman')
    except:
        # file might not exist
        return None
        
    # old pattern file format: odd lines are patterns, even lines are comments,
    # followed by (optionally) the active state (0/1)
    ret = []
    isComment = False
    idx = 0
    for line in f:
        line = line.strip()
        if isComment:
            isComment = False
            row = '<td>%s</td><td>%s</td>' % (line[1:line.rfind('"')], pat)
            if idx % 2:
                row = '<tr>%s</tr>' % row
            else:
                row = '<tr class="grayRow">%s</tr>' % row
                
            idx += 1
            ret.append(row)
        else:            
            pat = line[1:-1]
            isComment = True

    return ret

def backupOldSettings (srcPath, destPath):
    """Processes each configuration file of JM 1.19c/earlier, and put them in folder
    destPath/Old JunkMatcher Settings."""
    destPath = os.path.join(destPath, 'Old JunkMatcher Patterns')
    if os.path.exists(destPath):
        shutil.rmtree(destPath)
    os.mkdir(destPath)
    shutil.copy(ROOT_PATH + '.base.css', destPath)

    for srcFN, title, instruction in patternFiles:
        try:
            patternList = readPatternsFromFN(os.path.join(srcPath, srcFN))
            if instruction is None:
                # special cases
                if srcFN == 'metaPatterns.txt':
                    instruction = 'Please select the meta patterns you want to keep and copy/paste them into the Meta Patterns drawer in the new JunkMatcher.app.'
                elif srcFN == 'safeIP.txt':
                    instruction = 'Please select the IP patterns you want to keep, and copy/paste them into the "Safe IP Patterns" window - this window is accessible via the Test Inspector window over the property "Open relay" in the new JunkMatcher.app.'
                elif srcFN == 'safeRecipients.txt':
                    instruction = 'Please select the email address patterns that are used to send emails to you, and copy/paste them into the "Recipient Patterns" window - this window is accessible via the Test Inspector window over the property "Recipient(s) mismatch" in the new JunkMatcher.app.'
            else:
                instruction = 'The following patterns were used to match message view: "%s". \
Select the patterns you need and copy/paste them to target the same message view in the new JunkMatcher.app.' % instruction
        
            if patternList is not None:
                f = openFile(os.path.join(destPath, title + '.html'), 'w', 'utf8')            
                f.write(htmlTemplate % (title, 'Old patterns: %s' % title,
                                        '<b>Instruction:</b> %s' % instruction,
                                        '<div id="patternTable">\n<table>\n<tr><td><b>Pattern Names</b></td><td><b>Patterns</b></td></tr>%s</table>\n</div>' % '\n'.join(patternList)))

        except Exception, e:
            printException('Exception when backing up %s' % srcFN, e)

    # whitelist
    try:
        srcFN = os.path.join(srcPath, 'safeSenders.txt')
        if os.path.exists(srcFN):
            f = openFile(srcFN, 'r', 'macroman')
            patternList = []
            for idx, l in enumerate(f.readlines()):
                if idx % 2:
                    patternList.append('<tr><td>%s</td></tr>' % l.strip())
                else:
                    patternList.append('<tr class="grayRow"><td>%s</td></tr>' % l.strip())
            f.close()
        
            f = openFile(os.path.join(destPath, 'Whitelist.html'), 'w', 'utf8')
            f.write(htmlTemplate % (title, 'Old Whitelist',
                                    '<b>Instruction:</b> Please select the email addresses you want to whitelist into the Whitelist window of the new JunkMatcher.app (accessible via one of the toolbars).',
                                    '<div id="patternTable">\n<table>\n<tr><td><b>Email Addresses</b></td></tr>%s</table>\n</div>' % '\n'.join(patternList)))

    except Exception, e:
        printException('Exception when backing up safeSenders.txt', e)

    # safe sites
    try:
        srcFN = os.path.join(srcPath, 'safeSites.txt')
        if os.path.exists(srcFN):
            f = openFile(srcFN, 'r', 'macroman')
            patternList = []
            for idx, l in enumerate(f.readlines()):
                if idx % 2:
                    patternList.append('<tr><td>%s</td></tr>' % l.strip())
                else:
                    patternList.append('<tr class="grayRow"><td>%s</td></tr>' % l.strip())
            f.close()
        
            f = openFile(os.path.join(destPath, 'Safe Sites.html'), 'w', 'utf8')
            f.write(htmlTemplate % (title, 'Old list of Safe Sites',
                                    '<b>Instruction:</b> Please select the URL patterns you want to avoid collecting and copy/paste them into the Safe Sites drawer in the new JunkMatcher.app (accessible via the Sites window, on one of the toolbars).',
                                    '<div id="patternTable">\n<table>\n<tr><td><b>Email Addresses</b></td></tr>%s</table>\n</div>' % '\n'.join(patternList)))

    except Exception, e:
        printException('Exception when backing up safeSites.txt', e)


if __name__ == '__main__':
    backupOldSettings('/Users/ben/Public/JunkMatcher.mailbundle/Contents/Resources/junkMatcher/conf',
                      '/Users/ben/Desktop')
